
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"


/*------------------------------------------------------------------*/

void hscal_pctrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   hscalcomplex *a, *x;
 
   int num_of_processors = HMPI_Group_size(HMPI_COMM_WORLD_GROUP);
   int reconn            = N/sqrt(num_of_processors);
   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );

   struct timeval start, end;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&start, NULL);
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "UPLO %s TRANS %s DIAG %s N %d reconn %d IA %d JA %d IX %d JX %d "
           "INCX %d ",
           UPLO, TRANS, DIAG, N, reconn, IA, JA, IX, JX, INCX
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   x = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (RECONX)
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   //
   // Initialize the matrix A and vector x
   for (i = 0; i < reconn*reconn; i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   //
   // Call the BLAS ctrmv routine
   hscal_ctrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     a,
     &lda,
     x,
     &INCX
   );

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&end, NULL);

      {
         double tstart  = start.tv_sec + start.tv_usec/1000000.;
         double tend    = end.tv_sec + end.tv_usec/1000000.;
         double t_sec   = tend - tstart;

         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            t_sec
         );
      }
   }

   return;
}

/*------------------------------------------------------------------*/

int hscal_pack_pctrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pctrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pctrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon(
               &hscal_pctrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pctrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pctrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_d_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

void hscal_pdtrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   double *a, *x;
 
   int num_of_processors = HMPI_Group_size(HMPI_COMM_WORLD_GROUP);
   int reconn            = N/sqrt(num_of_processors);
   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );

   struct timeval start, end;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&start, NULL);
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "UPLO %s TRANS %s DIAG %s N %d reconn %d IA %d JA %d IX %d JX %d "
                  "INCX %d ",
                  UPLO, TRANS, DIAG, N, reconn, IA, JA, IX, JX,
                  INCX
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (double*)malloc(
                sizeof(double)
                *
                (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   x = (double*)malloc(
                sizeof(double)
                *
                RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   //
   // Initialize the matrix A and vector x
   for (i = 0; i < (reconn*reconn); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   //
   // Call the BLAS dtrmv routine
   hscal_dtrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     a,
     &lda,
     x,
     &INCX
   );

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&end, NULL);

      {
         double tstart  = start.tv_sec + start.tv_usec/1000000.;
         double tend    = end.tv_sec + end.tv_usec/1000000.;
         double t_sec   = tend - tstart;

         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            t_sec
         );
      }
   }

   return;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdtrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdtrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdtrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon(
               &hscal_pdtrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdtrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_d_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

void hscal_pstrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   float *a, *x;
 
   int num_of_processors = HMPI_Group_size(HMPI_COMM_WORLD_GROUP);
   int reconn            = N/sqrt(num_of_processors);
   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );

   struct timeval start, end;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&start, NULL);
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "UPLO %s TRANS %s DIAG %s N %d reconn %d IA %d JA %d IX %d JX %d "
                  "INCX %d ",
                  UPLO, TRANS, DIAG, N, reconn, IA, JA, IX, JX,
                  INCX
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (float*)malloc(
               sizeof(float)
               *
               (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   x = (float*)malloc(
               sizeof(float)
               *
               RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   //
   // Initialize the matrix A and vector x
   for (i = 0; i < (reconn*reconn); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   //
   // Call the BLAS strmv routine
   hscal_strmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     a,
     &lda,
     x,
     &INCX
   );

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&end, NULL);

      {
         double tstart  = start.tv_sec + start.tv_usec/1000000.;
         double tend    = end.tv_sec + end.tv_usec/1000000.;
         double t_sec   = tend - tstart;

         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            t_sec
         );
      }
   }

   return;
}

/*------------------------------------------------------------------*/

int hscal_pack_pstrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pstrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pstrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon(
               &hscal_pstrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pstrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pstrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_d_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

void hscal_pztrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   hscaldoublecomplex *a, *x;
 
   int num_of_processors = HMPI_Group_size(HMPI_COMM_WORLD_GROUP);
   int reconn            = N/sqrt(num_of_processors);
   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );

   struct timeval start, end;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&start, NULL);
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "UPLO %s TRANS %s DIAG %s N %d IA %d JA %d IX %d JX %d "
                  "INCX %d ",
                  UPLO, TRANS, DIAG, N, IA, JA, IX, JX,
                  INCX
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   x = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   //
   // Initialize the matrix A and vector x
   for (i = 0; i < (reconn*reconn); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   //
   // Call the BLAS ztrmv routine
   hscal_ztrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     a,
     &lda,
     x,
     &INCX
   );

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&end, NULL);

      {
         double tstart  = start.tv_sec + start.tv_usec/1000000.;
         double tend    = end.tv_sec + end.tv_usec/1000000.;
         double t_sec   = tend - tstart;

         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            t_sec
         );
      }
   }

   return;
}

/*------------------------------------------------------------------*/

int hscal_pack_pztrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pztrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +     
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *UPLO;
   (*model_params)[1]                                           = *TRANS;
   (*model_params)[1+1]                                         = *DIAG;
   (*model_params)[1+1+1]                                       = *N;
   (*model_params)[1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                           = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                         = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pztrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon(
               &hscal_pztrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pztrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pztrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_d_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
