
#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_pilaenv
(
    char *precision
)
{
    int ret_val;

/*  Purpose */
/*  ======= */

/*  This function  returns  the  positive integer value of the logical blocking */
/*  size. This value is machine and precision specific. This version pro- */
/*  vides a logical blocking size which should give good but  not optimal */
/*  performance on many  of the  currently  available  distributed-memory */
/*  concurrent computers.  Users are encouraged to modify this subroutine */
/*  to set this tuning parameter for their particular machine. */

/*  This function returns the algorithmic blocking factor and not the data */
/*  distribution blocking factor */

/*  Arguments */
/*  ========= */

/*  PREC    (global input) CHARACTER*1 */
/*          On input, PREC specifies the precision for which the  logical */
/*          block size should be returned as follows: */
/*             PREC = 'S' or 's' single precision real, */
/*             PREC = 'D' or 'd' double precision real, */
/*             PREC = 'C' or 'c' single precision complex, */
/*             PREC = 'Z' or 'z' double precision complex, */
/*             PREC = 'I' or 'i' integer. */

/*  Notes */
/*  ===== */

/*  Before modifying this routine to tune the library performance on your */
/*  system, be aware of the following: */

/*  1) The value this function returns must be STRICTLY LARGER THAN ZERO, */

/*  2) If you are planning to link your program with different  instances */
/*  of the library, (for example on a heterogeneous  machine),  you  MUST */
/*  compile each instance of the library with the  EXACT SAME  version of */
/*  this routine for obvious inter-operability reasons. */

/*  3) For performance models, provide a very large value as estimations  */
/*     are independent of this value                                      */

/*  ===================================================================== */

/*     .. External Functions .. */
/*     .. */
/*     .. Executable Statements .. */

    if (hscal_lsame(precision, "S")) {

/*        Single precision real logical block size */

        ret_val = 1024;

    } else if (hscal_lsame(precision, "D")) {

/*        Double precision real logical block size */

        ret_val = 1024;

    } else if (hscal_lsame(precision, "C")) {

/*        Single precision complex logical block size */

        ret_val = 1024;

    } else if (hscal_lsame(precision, "Z")) {

/*        Double precision complex logical block size */

        ret_val = 1024;

    } else if (hscal_lsame(precision, "I")) {

/*        Integer logical block size */

        ret_val = 1024;

    } else {

/*        Probably unused */

        ret_val = 1024;

    }

    return ret_val;

/*     End of hscal_pilaenv */

}

/*------------------------------------------------------------------*/
