
#include <math.h>
#include <stdarg.h> 
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include <hscalapack.h>

#include "hscalapack_pdgemv_parameters.h"

/*------------------------------------------------------------------*/

int Execute_benchmark_function
(
   int TRANSI,
   int M,
   int N,
   hscaldoublereal ALPHA,
   int INCX,
   hscaldoublereal BETA,
   int INCY
)
{
   int i;

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   double *a, *x, *y;

   //
   // Allocate the matrix A, vectors x and y
   a = (double*)malloc(
                sizeof(double)
                *
                (M*N)
   );

   if (a == NULL)
   {
      printf("Cannot allocate a\n");
      return HSCAL_ERR_LOW_MEMORY;
   }

   x = (double*)malloc(
                sizeof(double)
                *
                (N)
   );

   if (x == NULL)
   {
      free(a);
      printf("Cannot allocate x\n");
      return HSCAL_ERR_LOW_MEMORY;
   }

   y = (double*)malloc(
                sizeof(double)
                *
                (M)
   );


   if (y == NULL)
   {
      free(a);
      free(x);
      printf("Cannot allocate y\n");
      return HSCAL_ERR_LOW_MEMORY;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < (M*N); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < N; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < M; i++)
   {
      y[i] = 0.;
   }

   //
   // Call the BLAS dgemv routine
   hscal_dgemv_benchmark(
     TRANS,
     &M,
     &N,
     &ALPHA,
     a,
     &M,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    int        i, j, k, rc, plen, pnum, gsize, gpsize, me, parent_in_each_group = 0,
               optimal_number_of_processes = 0, proc_table_size;
    char       pname[MPI_MAX_PROCESSOR_NAME];
    MPI_Comm * level2comm;

    /*
    * Initialize HMPI runtime
    */
    HMPI_Init(&argc, &argv);

    rc = MPI_Get_processor_name(
            pname,
            &plen
    );

    if (rc != MPI_SUCCESS)
    {
       printf(
         "Problems with MPI_Get_processor_name... Aborting...\n"
       );
       HMPI_Abort(-1);
    }

    /* 
    * Each of the processes in HMPI_COMM_WORLD sends
    * its processor name using MPI_Get_processor_name to the host.
    * The host allocates a unique integer to each processor name
    * and sends the integer to the process launched on that
    * particular processor.
    */
    {
       int  *counts, *displs,
            *proc_array;

       char *proc_str, (*proc_table)[MPI_MAX_PROCESSOR_NAME];

       rc = MPI_Comm_size(
               HMPI_COMM_WORLD,
               &gsize
       );

       if (rc != MPI_SUCCESS)
       {
          printf(
            "Problems with MPI_Comm_size for HMPI_COMM_WORLD... Aborting...\n"
          );
          HMPI_Abort(-2);
       }

       rc = MPI_Comm_rank(
               HMPI_COMM_WORLD,
               &me
       );

       if (rc != MPI_SUCCESS)
       {
          printf(
            "Problems with MPI_Comm_rank for HMPI_COMM_WORLD... Aborting...\n"
          );
          HMPI_Abort(-3);
       }

       if (me == 0)
       {
          proc_str = (char*)malloc(
                            sizeof(char)
                            *
                            (MPI_MAX_PROCESSOR_NAME)*(gsize)
          );

          if (proc_str == NULL)
          {
             printf(
               "cant allocate proc_str...Low memory...Aborting...\n"
             );
             HMPI_Abort(-4);
          }

          proc_table = (char (*) [MPI_MAX_PROCESSOR_NAME])malloc(
                       sizeof(char[MPI_MAX_PROCESSOR_NAME])
                       *
                       gsize
          );

          if (proc_table == NULL)
          {
             printf(
               "Cant allocate proc_table...Low memory...Aborting...\n"
             );
             HMPI_Abort(-5);
          }

          proc_array = (int*)malloc(
                             sizeof(int)
                             *
                             gsize
          );

          if (proc_array == NULL)
          {
             printf(
               "Cant allocate proc_array...Low memory...Aborting...\n"
             );
             HMPI_Abort(-6);
          }

          counts = (int*)malloc(
                         sizeof(int)
                         *
                         gsize
          );

          if (counts == NULL)
          {
             printf(
               "Cant allocate counts...Low memory...Aborting...\n"
             );
             HMPI_Abort(-7);
          }

          displs = (int*)malloc(
                         sizeof(int)
                         *
                         gsize
          );

          if (displs == NULL)
          {
             printf(
               "Cant allocate displs...Low memory...Aborting...\n"
             );
             HMPI_Abort(-8);
          }

          for (i = 0; i < gsize; i++)
          {
             counts[i] = MPI_MAX_PROCESSOR_NAME;
             displs[i] = (i*MPI_MAX_PROCESSOR_NAME);
          }

          rc = MPI_Gatherv(
                  &pname,
                  MPI_MAX_PROCESSOR_NAME,
                  MPI_CHAR,
                  proc_str,
                  counts,
                  displs,
                  MPI_CHAR,
                  me,
                  HMPI_COMM_WORLD
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Gatherv for HMPI_COMM_WORLD... Aborting...\n"
             );
             HMPI_Abort(-9);
          }
       }
       else
       {
          rc = MPI_Gatherv(
                  &pname,
                  MPI_MAX_PROCESSOR_NAME,
                  MPI_CHAR,
                  NULL,
                  NULL,
                  NULL,
                  MPI_CHAR,
                  0,
                  HMPI_COMM_WORLD
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Gatherv for HMPI_COMM_WORLD... Aborting...\n"
             );
             HMPI_Abort(-9);
          }
       }

       if (me == 0)
       {
          strncpy(
             proc_table[0],
             proc_str,
             MPI_MAX_PROCESSOR_NAME
          );

          proc_table_size = 1;
          for (i = 0; i < gsize; i++)
          {
             int entry_found = 0;
             char* tmp = proc_str
                         +
                         (i)*(MPI_MAX_PROCESSOR_NAME)
             ;

             for (j = 0; j < proc_table_size; j++)
             {
                 if (strncmp(
                        tmp,
                        proc_table[j],
                        MPI_MAX_PROCESSOR_NAME
                    )
                    == 0
                 )
                 {
                    entry_found = 1;
                    break;
                 }
             }

             if (!entry_found)
             {
                proc_table_size++;

                strncpy(
                   proc_table[(proc_table_size - 1)],
                   tmp,
                   MPI_MAX_PROCESSOR_NAME
                );
             }
          }

          for (i = 0; i < gsize; i++)
          {
             char* tmp = proc_str
                         +
                         (i)*(MPI_MAX_PROCESSOR_NAME)
             ;

             for (j = 0; j < proc_table_size; j++)
             {
                if (strncmp(
                       tmp,
                       proc_table[j],
                       MPI_MAX_PROCESSOR_NAME
                   )
                   == 0
                )
                {
                   proc_array[i] = j;
                }
             }
          }

          for (i = 0; i < gsize; i++)
          {
             counts[i] = 1;
             displs[i] = i;
          }

          rc = MPI_Scatterv(
                  proc_array,
                  counts,
                  displs,
                  MPI_INT,
                  &pnum,
                  1,
                  MPI_INT,
                  0,
                  HMPI_COMM_WORLD
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Gatherv for HMPI_COMM_WORLD... Aborting...\n"
             );
             HMPI_Abort(-10);
          }
       }
       else
       {
          rc = MPI_Scatterv(
                  NULL,
                  NULL,
                  NULL,
                  MPI_INT,
                  &pnum,
                  1,
                  MPI_INT,
                  0,
                  HMPI_COMM_WORLD
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Gatherv for HMPI_COMM_WORLD... Aborting...\n"
             );
             HMPI_Abort(-10);
          }
       }

       if (me == 0)
       {
          free(counts);
          free(displs);
          free(proc_str);
          free(proc_table);
          free(proc_array);
       }
    }

    /* 
    * Broadcast the number of computers
    */
    rc = MPI_Bcast(
            &proc_table_size,
            1,
            MPI_INT,
            0,
            HMPI_COMM_WORLD
    );

    if (rc != MPI_SUCCESS)
    {
       printf(
         "Problems with MPI_Bcast for HMPI_COMM_WORLD... Aborting...\n"
       );
       HMPI_Abort(-25);
    }

    level2comm = (MPI_Comm*)malloc(
                            sizeof(MPI_Comm)
                            *
                            proc_table_size
    );

    if (level2comm == NULL)
    {
       printf(
         "Cant allocate an array of level2comm...Low memory...Aborting...\n"
       );
       HMPI_Abort(-27);
    }

    for (i = 0; i < proc_table_size; i++)
    {
       int color = MPI_UNDEFINED;

       /*
       * A communicator for each processor in the communication universe
       */
       if (i == pnum)
       {
          color = 0;
       }

       rc = MPI_Comm_split(
               HMPI_COMM_WORLD,
               color,
               me,
               &level2comm[i]
       );

       if (rc != MPI_SUCCESS)
       {
          printf(
             "Problems with MPI_Comm_split for HMPI_COMM_WORLD... Aborting...\n"
          );
          HMPI_Abort(-11);
       }
    }

    for (i = 0; i < proc_table_size; i++)
    {
       /*
       * Now for one or more processes on the processor
       */
       if (i == pnum)
       {
          int psize, me_s;

          rc = MPI_Comm_size(
                  level2comm[i],
                  &psize
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Comm_size for level2comm... Aborting...\n"
             );
             HMPI_Abort(-12);
          }

          rc = MPI_Comm_rank(
                  level2comm[i],
                  &me_s
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Comm_rank for level2comm... Aborting...\n"
             );
             HMPI_Abort(-13);
          }

          if (me_s == 0)
          {
             parent_in_each_group = 1;
          }

          for (j = 0; j < psize; j++)
          {
             double maxtprev;
             MPI_Comm level2comm_s;
             int color_s = 0;
             int breakout = 0;

             if (me_s > j)
             {
                color_s = MPI_UNDEFINED;
             } 

             rc = MPI_Comm_split(
                     level2comm[i],
                     color_s,
                     me_s,
                     &level2comm_s
             );

             if (rc != MPI_SUCCESS)
             {
                printf(
                  "Problems with MPI_Comm_split for level2comm... Aborting...\n"
                );
                HMPI_Abort(-14);
             }

             /*
             * Processes on the processor execute the benchmark code
             * as follows
             *
             * 0             only process 0
             * 0 1           processes 0 and 1 in parallel
             * 0 1 2         processes 0, 1, and 2 in parallel
             * 0 1 2 3       processes 0, 1, 3, and 4 in parallel
             * ...
             *
             * Peformance degradation with addition of a process is 
             * calculated
             */
             if (me_s > j)
             {
                /*
                * Wait for signal to continue or break
                */
                rc = MPI_Bcast(
                        &breakout,
                        1,
                        MPI_INT,
                        0,
                        level2comm[i]
                );

                if (rc != MPI_SUCCESS)
                {
                   printf(
                     "Problems with MPI_Bcast for level2comm_s... Aborting...\n"
                   );
                   HMPI_Abort(-26);
                }

                if (breakout)
                {
                   break;
                }

                MPI_Barrier(level2comm[i]);

                continue;                  
             }
             else
             {
                double start, end, etimeavg, etimecum = 0.;
                int *counts_s, *displs_s;
                int psize_s, me_s_s;
                int parent = 0;
                int low_memory = 0;

                /*
                * Now execute the benchmark code
                */
                for (k = 0; k < NUMBER_OF_BENCH_RUNS; k++)
                {
                   start = MPI_Wtime();

                   rc = Execute_benchmark_function(
                          TRANSI,
                          M,
                          N,
                          ALPHA,
                          INCX,
                          BETA,
                          INCY
                   );                

                   if (rc == HSCAL_ERR_LOW_MEMORY)
                   {
                      low_memory = 1;
                      break;
                   }

                   end = MPI_Wtime();

                   etimecum += end - start;
                }

                /*
                * Gather low memory flag
                */
                rc = MPI_Comm_rank(
                        level2comm_s,
                        &me_s_s
                );

                if (rc != MPI_SUCCESS)
                {
                   printf(
                     "Problems with MPI_Comm_rank for level2comm_s... Aborting...\n"
                   );
                   HMPI_Abort(-15);
                }

                rc = MPI_Comm_size(
                        level2comm_s,
                        &psize_s
                );

                if (rc != MPI_SUCCESS)
                {
                   printf(
                     "Problems with MPI_Comm_size for level2comm_s... Aborting...\n"
                   );
                   HMPI_Abort(-16);
                }

                if (me_s_s == parent)
                {
                   int *lmemoryarray = (int*)malloc(
                                             sizeof(int)
                                             *
                                             psize_s
                   );

                   if (lmemoryarray == NULL)
                   {
                      printf(
                        "Cant allocate lmemoryarray ...Low memory...Aborting...\n"
                      );
                      HMPI_Abort(-17);
                   }

                   counts_s = (int*)malloc(
                                    sizeof(int)
                                    *
                                    psize_s
                   );

                   if (counts_s == NULL)
                   {
                      printf(
                        "Cant allocate counts_s...Low memory...Aborting...\n"
                      );
                      free(lmemoryarray);
                      HMPI_Abort(-18);
                   }

                   displs_s = (int*)malloc(
                                    sizeof(int)
                                    *
                                    psize_s
                   );

                   if (displs_s == NULL)
                   {
                      printf(
                        "Cant allocate displs_s...Low memory...Aborting...\n"
                      );
                      free(counts_s);
                      free(lmemoryarray);
                      HMPI_Abort(-19);
                   }

                   for (k = 0; k < psize_s; k++)
                   {
                      counts_s[k] = 1;
                      displs_s[k] = k;
                   }

                   rc = MPI_Gatherv(
                           &low_memory,
                           1,
                           MPI_INT,
                           lmemoryarray,
                           counts_s,
                           displs_s,
                           MPI_INT,
                           parent,
                           level2comm_s
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Gatherv for level2comm_s... Aborting...\n"
                      );
                      HMPI_Abort(-20);
                   }
               
                   free(counts_s);
                   free(displs_s);

                   for (k = 0; k < psize_s; k++)
                   {
                      if (lmemoryarray[k] == 1)
                      {
                         breakout = 1;
                         break;
                      }
                   }

                   rc = MPI_Bcast(
                           &breakout,
                           1,
                           MPI_INT,
                           0,
                           level2comm_s
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Bcast for level2comm_s... Aborting...\n"
                      );
                      HMPI_Abort(-26);
                   }

                   if (breakout)
                   {
                      rc = MPI_Bcast(
                              &breakout,
                              1,
                              MPI_INT,
                              0,
                              level2comm[i]
                      );

                      if (rc != MPI_SUCCESS)
                      {
                         printf(
                           "Problems with MPI_Bcast for level2comm... Aborting...\n"
                         );
                         HMPI_Abort(-26);
                      }
                   }

                   free(lmemoryarray);
                }
                else 
                {
                   rc = MPI_Gatherv(
                           &low_memory,
                           1,
                           MPI_INT,
                           NULL,
                           NULL,
                           NULL,
                           MPI_INT,
                           parent,
                           level2comm_s
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Gatherv for level2comm_s... Aborting...\n"
                      );
                      HMPI_Abort(-20);
                   }

                   rc = MPI_Bcast(
                           &breakout,
                           1,
                           MPI_INT,
                           0,
                           level2comm_s
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Bcast for level2comm_s... Aborting...\n"
                      );
                      HMPI_Abort(-26);
                   }

                   if (breakout)
                   {
                      rc = MPI_Bcast(
                              &breakout,
                              1,
                              MPI_INT,
                              0,
                              level2comm[i]
                      );

                      if (rc != MPI_SUCCESS)
                      {
                         printf(
                           "Problems with MPI_Bcast for level2comm... Aborting...\n"
                         );
                         HMPI_Abort(-26);
                      }
                   }
                }

                if (breakout)
                {
                   rc = MPI_Comm_free(&level2comm_s);

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "MAIN:Problems freeing level2comm_s"
                        "...Aborting...\n"
                      );

                      HMPI_Abort(-9);
                   } 

                   break;
                }

                /*
                * Seems we have memory
                * So proceed with the VPM determination
                */
                etimeavg = etimecum/((double)NUMBER_OF_BENCH_RUNS);

                if (me_s_s == parent)
                {
                   double maxt, perfdgd;
                   double *tarray = (double*)malloc(
                                             sizeof(double)
                                             *
                                             psize_s
                   );

                   if (tarray == NULL)
                   {
                      printf(
                        "Cant allocate tarray...Low memory...Aborting...\n"
                      );
                      HMPI_Abort(-17);
                   }

                   counts_s = (int*)malloc(
                                    sizeof(int)
                                    *
                                    psize_s
                   );

                   if (counts_s == NULL)
                   {
                      printf(
                        "Cant allocate counts_s...Low memory...Aborting...\n"
                      );
                      HMPI_Abort(-18);
                   }

                   displs_s = (int*)malloc(
                                    sizeof(int)
                                    *
                                    psize_s
                   );

                   if (displs_s == NULL)
                   {
                      printf(
                        "Cant allocate displs_s...Low memory...Aborting...\n"
                      );
                      HMPI_Abort(-19);
                   }

                   for (k = 0; k < psize_s; k++)
                   {
                      counts_s[k] = 1;
                      displs_s[k] = k;
                   }

                   rc = MPI_Gatherv(
                           &etimeavg,
                           1,
                           MPI_DOUBLE,
                           tarray,
                           counts_s,
                           displs_s,
                           MPI_DOUBLE,
                           parent,
                           level2comm_s
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Gatherv for level2comm_s... Aborting...\n"
                      );
                      HMPI_Abort(-20);
                   }
               
                   /*
                   * Now evaluate the performance degradation
                   * First get the maximum
                   */
                   if (psize_s == 1)
                   {
                      maxtprev = tarray[0];
                   }

                   maxt = tarray[0];

                   for (k = 0; k < psize_s; k++)
                   {
                      if (tarray[k] > maxt)
                      {
                         maxt = tarray[k];
                      }
                   }

                   perfdgd = ((maxt - maxtprev)/maxtprev)*100.00;

                   if (HSCAL_TEST_LOG_VERBOSE)
                   {
                      printf("%s: Performance degradation %f Times are ", pname, perfdgd);

                      for (k = 0; k < psize_s; k++)
                      {
                         printf("%f ", tarray[k]);
                      }
                      printf("\n");
                   }

                   /*
                   * Send a signal to the processes to quit
                   * if the performance degradation is too large
                   */
                   if ((perfdgd > 0)
                       && (perfdgd >= MAX_PERF_DGRD_ALLOWED
                      )
                   )
                   {
                      breakout = 1;
                   }
                   else
                   {
                      maxtprev = maxt;
                      optimal_number_of_processes = psize_s;
                   }

                   rc = MPI_Bcast(
                           &breakout,
                           1,
                           MPI_INT,
                           0,
                           level2comm[i]
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Bcast for level2comm... Aborting...\n"
                      );
                      HMPI_Abort(-26);
                   }

                   free(counts_s);
                   free(displs_s);
                   free(tarray);
                }
                else 
                {
                   rc = MPI_Gatherv(
                           &etimeavg,
                           1,
                           MPI_DOUBLE,
                           NULL,
                           NULL,
                           NULL,
                           MPI_DOUBLE,
                           parent,
                           level2comm_s
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Gatherv for level2comm_s... Aborting...\n"
                      );
                      HMPI_Abort(-20);
                   }

                   rc = MPI_Bcast(
                           &breakout,
                           1,
                           MPI_INT,
                           0,
                           level2comm[i]
                   );

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "Problems with MPI_Bcast for level2comm... Aborting...\n"
                      );
                      HMPI_Abort(-26);
                   }

                }

                if (breakout)
                {
                   rc = MPI_Comm_free(&level2comm_s);

                   if (rc != MPI_SUCCESS)
                   {
                      printf(
                        "MAIN:Problems freeing level2comm_s"
                        "...Aborting...\n"
                      );

                      HMPI_Abort(-9);
                   } 

                   break;
                }

                MPI_Barrier(level2comm[i]);
             }

             rc = MPI_Comm_free(&level2comm_s);

             if (rc != MPI_SUCCESS)
             {
                printf(
                  "MAIN:Problems freeing level2comm_s"
                  "...Aborting...\n"
                );

                HMPI_Abort(-9);
             } 
          }
       }
    }

    for (i = 0; i < proc_table_size; i++)
    {
       if (i == pnum)
       {
          rc = MPI_Comm_free(&level2comm[i]);

          if (rc != MPI_SUCCESS)
          {
             printf(
               "MAIN:Problems freeing level2comm "
               "...Aborting...\n"
             );

             HMPI_Abort(-28);
          }
       }
    }

    free(level2comm);
    
    /*
    * Now gather the results at the host process of the 
    * communication universe
    */
    {
       MPI_Comm resultscomm;
       int color = MPI_UNDEFINED;

       /*
       * Now for each processor in the communication universe
       */
       if (parent_in_each_group == 1)
       {
          color = 0;
       }

       rc = MPI_Comm_split(
               HMPI_COMM_WORLD,
               color,
               me,
               &resultscomm
       );

       if (rc != MPI_SUCCESS)
       {
          printf(
             "Problems with MPI_Comm_split for resultscomm... Aborting...\n"
          );
          HMPI_Abort(-21);
       }

       if (parent_in_each_group == 1)
       {
          /*
          * Gather the processor and the optimal
          * number of processes on it
          */
          char *proc_str;
          int *optn;
          int hsize;
          int hme;

          rc = MPI_Comm_size(
                  resultscomm,
                  &hsize
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Comm_size for resultscomm... Aborting...\n"
             );
             HMPI_Abort(-22);
          } 

          rc = MPI_Comm_rank(
                  resultscomm,
                  &hme
          );

          if (rc != MPI_SUCCESS)
          {
             printf(
               "Problems with MPI_Comm_rank for resultscomm... Aborting...\n"
             );
             HMPI_Abort(-23);
          }

          /*
          * Get the processor names
          */
          if (hme == 0)
          {
             int *counts, *displs;

             proc_str = (char*)malloc(
                               sizeof(char)
                               *
                               (MPI_MAX_PROCESSOR_NAME)*(hsize)
             );

             if (proc_str == NULL)
             {
                printf(
                  "cant allocate proc_str...Low memory...Aborting...\n"
                );
                HMPI_Abort(-24);
             }

             counts = (int*)malloc(
                            sizeof(int)
                            *
                            hsize
             );

             if (counts == NULL)
             {
                printf(
                  "Cant allocate counts...Low memory...Aborting...\n"
                );
                HMPI_Abort(-25);
             }

             displs = (int*)malloc(
                            sizeof(int)
                            *
                            hsize
             );

             if (displs == NULL)
             {
                printf(
                  "Cant allocate displs...Low memory...Aborting...\n"
                );
                HMPI_Abort(-26);
             }

             for (i = 0; i < hsize; i++)
             {
                counts[i] = MPI_MAX_PROCESSOR_NAME;
                displs[i] = (i*MPI_MAX_PROCESSOR_NAME);
             }

             rc = MPI_Gatherv(
                     &pname,
                     MPI_MAX_PROCESSOR_NAME,
                     MPI_CHAR,
                     proc_str,
                     counts,
                     displs,
                     MPI_CHAR,
                     hme,
                     resultscomm
             );

             if (rc != MPI_SUCCESS)
             {
                printf(
                  "Problems with MPI_Gatherv for resultscomm... Aborting...\n"
                );
                HMPI_Abort(-27);
             }

             free(counts);
             free(displs);
          }
          else
          {
             rc = MPI_Gatherv(
                     &pname,
                     MPI_MAX_PROCESSOR_NAME,
                     MPI_CHAR,
                     NULL,
                     NULL,
                     NULL,
                     MPI_CHAR,
                     0,
                     resultscomm
             );

             if (rc != MPI_SUCCESS)
             {
                printf(
                  "Problems with MPI_Gatherv for resultscomm... Aborting...\n"
                );
                HMPI_Abort(-27);
             }
          }

          /*
          * Get the number of processes on each processor
          */
          if (hme == 0)
          {
             int *counts, *displs;

             optn = (int*)malloc(
                          sizeof(int)
                          *
                          hsize
             );

             if (proc_str == NULL)
             {
                printf(
                  "cant allocate optn...Low memory...Aborting...\n"
                );
                HMPI_Abort(-28);
             }

             counts = (int*)malloc(
                            sizeof(int)
                            *
                            hsize
             );

             if (counts == NULL)
             {
                printf(
                  "Cant allocate counts...Low memory...Aborting...\n"
                );
                HMPI_Abort(-29);
             }

             displs = (int*)malloc(
                            sizeof(int)
                            *
                            hsize
             );

             if (displs == NULL)
             {
                printf(
                  "Cant allocate displs...Low memory...Aborting...\n"
                );
                HMPI_Abort(-30);
             }

             for (i = 0; i < hsize; i++)
             {
                counts[i] = 1;
                displs[i] = i;
             }

             rc = MPI_Gatherv(
                     &optimal_number_of_processes,
                     1,
                     MPI_INT,
                     optn,
                     counts,
                     displs,
                     MPI_INT,
                     hme,
                     resultscomm
             );

             if (rc != MPI_SUCCESS)
             {
                printf(
                  "Problems with MPI_Gatherv for resultscomm... Aborting...\n"
                );
                HMPI_Abort(-31);
             }

             free(counts);
             free(displs);
          }
          else
          {
             rc = MPI_Gatherv(
                     &optimal_number_of_processes,
                     1,
                     MPI_INT,
                     NULL,
                     NULL,
                     NULL,
                     MPI_INT,
                     0,
                     resultscomm
             );

             if (rc != MPI_SUCCESS)
             {
                printf(
                  "Problems with MPI_Gatherv for resultscomm... Aborting...\n"
                );
                HMPI_Abort(-31);
             }
          }

          /*
          * Print the VPM
          */
          if (hme == 0)
          {
             printf("==================================================\n");
             printf("Virtual Parallel Machine:\n");
             printf("<Machine name> <number of processes> <scalability>\n");
             printf("==================================================\n");

             for (i = 0; i < hsize; i++)
             {
                char *procname = proc_str
                                 +
                                 (i)*(MPI_MAX_PROCESSOR_NAME)
                ;

                printf("%s %d [%d]\n", procname, optn[i], optn[i]);
             }

             printf("==================================================\n");

             free(proc_str);
             free(optn);
          }

          rc = MPI_Comm_free(&resultscomm);

          if (rc != MPI_SUCCESS)
          {
             printf(
               "MAIN:Problems freeing resultscomm "
               "...Aborting...\n"
             );

             HMPI_Abort(-19);
          } 
       }
    }

    HMPI_Finalize(0);
}
