
/*******************************************************************
*                                                                  *
*                       Heterogeneous ScaLAPACK                    *
*                                TOOLS                             *
*                             (INTERNAL)                           *
*                 ====================================             *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_TOOLS_HH
#define __HMPI_SCALAPACK_TOOLS_HH

int hscal_pchk1mat(
         hscalinteger *ma,
         hscalinteger *mapos0,
         hscalinteger *na,
         hscalinteger *napos0,
         hscalinteger *ia,
         hscalinteger *ja,
         hscalinteger *desca,
         hscalinteger *descapos0,
         hscalinteger *nextra,
         hscalinteger *ex,
         hscalinteger *expos,
         hscalinteger *info,
         hscalinteger *myrow,
         hscalinteger *mycol,
         hscalinteger *nprow,
         hscalinteger *npcol
);

int hscal_pchk2mat(
         hscalinteger *ma,
         hscalinteger *mapos0,
         hscalinteger *na,
         hscalinteger *napos0,
         hscalinteger *ia,
         hscalinteger *ja,
         hscalinteger *desca,
         hscalinteger *descapos0,
         hscalinteger *mb,
         hscalinteger *mbpos0,
         hscalinteger *nb,
         hscalinteger *nbpos0,
         hscalinteger *ib,
         hscalinteger *jb,
         hscalinteger *descb,
         hscalinteger *descbpos0,
         hscalinteger *nextra,
         hscalinteger *ex,
         hscalinteger *expos,
         hscalinteger *info,
         hscalinteger *myrow,
         hscalinteger *mycol,
         hscalinteger *nprow,
         hscalinteger *npcol
);

int hscal_globchk(
         hscalinteger *ictxt,
         hscalinteger *n,
         hscalinteger *x,
         hscalinteger *ldx,
         hscalinteger *iwork,
         hscalinteger *info,
         hscalinteger *myrow,
         hscalinteger *mycol,
         hscalinteger *nprow,
         hscalinteger *npcol
);

int hscal_picol2row(
         hscalinteger *ictxt,
         hscalinteger *m,
         hscalinteger *n,
         hscalinteger *nb,
         hscalinteger *ldvs,
         hscalinteger *ldvd,
         hscalinteger *rsrc,
         hscalinteger *csrc,
         hscalinteger *rdest,
         hscalinteger *cdest,
         int           nprow,
         int           npcol,
         int           myrow,
         int           mycol,
         double        *tcomp,
         int           *tcomm
);

int hscal_pirow2col(
         hscalinteger *ictxt,
         hscalinteger *m,
         hscalinteger *n,
         hscalinteger *nb,
         hscalinteger *ldvs,
         hscalinteger *ldvd,
         hscalinteger *rsrc,
         hscalinteger *csrc,
         hscalinteger *rdest,
         hscalinteger *cdest,
         int nprow,
         int npcol,
         int myrow,
         int mycol,
         double *tcomp,
         int *tcomm
);

hscalinteger hscal_ilcm(
         hscalinteger *m, 
         hscalinteger *n
);

int hscal_ilacpy(
         char *uplo,
         hscalinteger *m,
         hscalinteger *n,
         hscalinteger *lda,
         hscalinteger *ldb,
         double *tcomp
);

#define ilacpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ilacpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_indicative_complex(
         hscalcomplex *hc,
         hscalinteger *ihc
);

int hscal_indicative_doublecomplex(
         hscaldoublecomplex *hdc,
         hscalinteger *ihdc
);

int hscal_chk1mat(
         hscalinteger *ma,
         hscalinteger *mapos0,
         hscalinteger *na,
         hscalinteger *napos0,
         hscalinteger *ia,
         hscalinteger *ja,
         hscalinteger *desca,
         hscalinteger *descapos0,
         hscalinteger *info,
         hscalinteger myrow,
         hscalinteger mycol,
         hscalinteger nprow,
         hscalinteger npcol
);

int hscal_cmatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_dmatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_smatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_zmatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_descinit(
    hscalinteger *desc, 
    hscalinteger *m,  
    hscalinteger *n, 
    hscalinteger *mb, 
    hscalinteger *nb, 
    hscalinteger *irsrc, 
    hscalinteger *icsrc, 
    hscalinteger *info
);

int hscal_pcdbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *laf, hscalinteger *lwork, hscalinteger *info, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pddbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *laf, hscalinteger *lwork, hscalinteger *info, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_psdbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *laf, hscalinteger *lwork, hscalinteger *info, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pzdbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *laf, hscalinteger *lwork, hscalinteger *info, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pclacgv(
    hscalinteger *n,  hscalinteger *ix, hscalinteger *jx, hscalinteger *descx, hscalinteger *incx,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pzlacgv(
    hscalinteger *n,  hscalinteger *ix, hscalinteger *jx, hscalinteger *descx, hscalinteger *incx,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

#endif

